import argparse
import datetime
import hashlib
import logging
from typing import Any, Dict, Optional, Text, Union, Collection

from rasa.cli import utils as rasa_cli_utils
from rasa.constants import GLOBAL_USER_CONFIG_PATH

from rasa.nlu.training_data import TrainingData as NluTrainingData
from rasa.utils import common as rasa_utils

from rasax.community import config
from rasax.community.constants import (
    CONFIG_METRICS_ENABLED,
    CONFIG_METRICS_USER,
    MIXPANEL_TOKEN,
    CONFIG_FILE_METRICS_KEY,
    CONFIG_METRICS_DATE,
)
from rasax.community.version import __version__

logger = logging.getLogger(__name__)


def _mixpanel():
    """Create a mixpanel API instance."""
    from mixpanel import Mixpanel

    return Mixpanel(MIXPANEL_TOKEN)


def are_metrics_enabled() -> bool:
    """Did the user consent to the metrics collection."""
    return config.metrics_collection_config.get(CONFIG_METRICS_ENABLED, False)


def metrics_id() -> Optional[Text]:
    """Random unique consistent id."""
    return config.metrics_collection_config.get(CONFIG_METRICS_USER, None)


def _safe_track(
    distinct_id: Optional[Text],
    event_name: Text,
    properties: Optional[Dict[Text, Any]] = None,
    meta: Optional[Dict[Text, Any]] = None,
):
    try:
        _mixpanel().track(distinct_id, event_name, properties, meta)
    except Exception as e:
        logger.debug("Skipping metrics collection. {}".format(e))


def track_metrics_consent(
    user_id: Text, allow_metrics: bool, auto_accept: bool = False
):
    """Track metric consent."""
    _safe_track(
        user_id,
        "Metrics Consent",
        {"allow_metrics_collection": allow_metrics, "auto_accept": auto_accept},
    )


def track(
    event_name: Text,
    properties: Optional[Dict[Text, Any]] = None,
    meta: Optional[Dict[Text, Any]] = None,
):
    """If enabled, track certain statistical metrics."""

    if not are_metrics_enabled():
        return

    if properties:
        properties["metrics_id"] = metrics_id()

    _safe_track(metrics_id(), event_name, properties, meta)


def track_local_start(
    project_path: Text,
    domain: Dict,
    nlu_data: NluTrainingData,
    stories: Collection[Any],
) -> None:
    """Tracks an event when the local Rasa X edition is started."""

    project = hashlib.md5(project_path.encode("utf-8")).hexdigest()
    start_event = _get_rasa_x_start_event(project, domain, nlu_data, stories)
    track("Local X Start", start_event)


def track_server_start(
    domain: Dict, nlu_data: NluTrainingData, stories: Collection[Any]
) -> None:
    """Tracks an event when the dockerized Rasa X edition is started."""

    start_event = _get_rasa_x_start_event(
        config.project_name, domain, nlu_data, stories
    )
    track("Server X Start", start_event)


def _get_rasa_x_start_event(
    project: Text, domain: Dict, nlu_data: NluTrainingData, stories: Collection[Any]
) -> Dict[Text, Union[int, Text]]:
    return {
        "project": project,
        "rasa_x": __version__,
        "num_intent_examples": len(nlu_data.intent_examples),
        "num_entity_examples": len(nlu_data.entity_examples),
        "num_actions": len(domain.get("actions", [])),
        "num_templates": len(domain.get("templates", [])),
        "num_slots": len(domain.get("slots", [])),
        "num_forms": len(domain.get("forms", [])),
        "num_intents": len(domain.get("intents", [])),
        "num_entities": len(domain.get("entities", [])),
        "num_stories": len(stories),
    }


def metrics_collection_config(args: argparse.Namespace) -> Dict[Text, Any]:
    """Make sure the user consents to any metrics collection."""

    import questionary

    try:
        return rasa_utils.read_global_config_value(
            CONFIG_FILE_METRICS_KEY, unavailable_ok=False
        )
    except ValueError:
        pass  # swallow the error and ask the user

    allow_metrics = (
        questionary.confirm(
            "Rasa will track a minimal amount of anonymized usage information "
            "(like how often you use the 'train' button) to help us improve Rasa X. "
            "None of your training data or conversations will ever be sent to Rasa. "
            "Are you OK with Rasa collecting anonymized usage data?"
        )
        .skip_if(args.no_prompt, default=True)
        .ask()
    )

    if not args.no_prompt:
        rasa_cli_utils.print_success(
            "Your decision has been stored into {}.".format(GLOBAL_USER_CONFIG_PATH)
        )

    settings = _metrics_settings(allow_metrics)

    rasa_utils.write_global_config_value(CONFIG_FILE_METRICS_KEY, settings)
    track_metrics_consent(settings[CONFIG_METRICS_USER], allow_metrics, args.no_prompt)
    return settings


def _metrics_settings(
    enable_metrics: bool,
    metrics_user_id: Optional[Text] = None,
    consent_date: Optional[datetime.datetime] = None,
) -> Dict[Text, Any]:
    import uuid

    return {
        CONFIG_METRICS_ENABLED: enable_metrics,
        CONFIG_METRICS_USER: metrics_user_id or uuid.uuid4().hex,
        CONFIG_METRICS_DATE: consent_date or datetime.datetime.now(),
    }


def set_metrics_config_from_env() -> None:
    """Set the metrics configuration based on environment variables."""

    config.metrics_collection_config = _metrics_settings(
        config.metrics_consent_given, config.metrics_user_id
    )
